@@ -8,13 +8,14 @@ module Squeal.QuasiQuotes.Insert (
   toSquealInsert,
 ) where
 
-import Control.Monad (MonadFail(fail), mapM, when, zipWithM)
-import Data.Foldable (Foldable(foldr, length), foldlM)
+import Control.Monad (when, zipWithM)
+import Data.Foldable (foldlM)
 import Data.Maybe (isJust)
 import Language.Haskell.TH.Syntax (Exp(AppE, ConE, LabelE, ListE, VarE), Q)
 import Prelude
-  ( Applicative(pure), Either(Left), Eq((/=)), Maybe(Just, Nothing)
-  , Semigroup((<>)), Show(show), ($), (.), error, otherwise
+  ( Applicative(pure), Either(Left), Eq((/=)), Foldable(foldr, length)
+  , Maybe(Just, Nothing), MonadFail(fail), Semigroup((<>)), Show(show)
+  , Traversable(mapM), ($), (.), (<$>), error, otherwise
   )
 import Squeal.QuasiQuotes.Query (getIdentText, renderPGTAExpr, toSquealQuery)
 import qualified Data.List.NonEmpty as NE
@@ -77,62 +78,58 @@ toSquealInsert
           (names, exp) <- renderPGTWithClause withClause
           pure (names, Just exp)
 
-    when (isJust maybeOnConflict) $
-      fail "ON CONFLICT clauses are not supported yet."
+    let
+      table = renderPGTInsertTarget insertTarget
 
-    insertBody <-
+    queryClauseExp <-
       case insertRest of
         PGT_AST.SelectInsertRest maybeInsertColumnList maybeOverrideKind selectStmt -> do
           when (isJust maybeOverrideKind) $
             fail "OVERRIDING clause is not supported yet."
-          queryClauseExp <-
-            case selectStmt of
-              -- Case 1: INSERT ... VALUES ...
-              Left
-                (PGT_AST.SelectNoParens _ (Left (PGT_AST.ValuesSimpleSelect valuesClause)) _ _ _) ->
-                  case maybeInsertColumnList of
-                    Just colItems ->
-                      renderPGTValueRows cteNames (NE.toList colItems) valuesClause
-                    Nothing ->
-                      fail
-                        "INSERT INTO ... VALUES must specify column names for the Squeal-QQ translation."
-              -- Case 2: INSERT ... SELECT ... (a general SELECT statement)
-              _ ->
-                -- selectStmt is not a ValuesSimpleSelect (i.e., it's a general query)
+          case selectStmt of
+            -- Case 1: INSERT ... VALUES ...
+            Left
+              (PGT_AST.SelectNoParens _ (Left (PGT_AST.ValuesSimpleSelect valuesClause)) _ _ _) ->
                 case maybeInsertColumnList of
-                  Just _ ->
+                  Just colItems ->
+                    renderPGTValueRows cteNames (NE.toList colItems) valuesClause
+                  Nothing ->
                     fail
-                      "INSERT INTO table (columns) SELECT ... is not yet supported by Squeal-QQ. Please use INSERT INTO table SELECT ... and ensure your SELECT statement provides all columns for the table, matching the table's column order and types."
-                  Nothing -> do
-                    squealQueryExp <- toSquealQuery cteNames Nothing selectStmt -- from Squeal.QuasiQuotes.Query
-                    pure (ConE 'S.Subquery `AppE` squealQueryExp)
-          let
-            table = renderPGTInsertTarget insertTarget
-          case maybeReturningClause of
-            Nothing ->
-              pure $ VarE 'S.insertInto_ `AppE` table `AppE` queryClauseExp
-            Just (NE.toList -> [PGT_AST.AsteriskTargetEl]) -> do
-              let
-                returning = ConE 'S.Returning `AppE` ConE 'S.Star
-              pure $
-                VarE 'S.insertInto
-                  `AppE` table
-                  `AppE` queryClauseExp
-                  `AppE` ConE 'S.OnConflictDoRaise
-                  `AppE` returning
-            Just targetList -> do
-              returningProj <- renderTargetList cteNames (NE.toList targetList)
-              let
-                returning = ConE 'S.Returning `AppE` (ConE 'S.List `AppE` returningProj)
-              pure $
-                VarE 'S.insertInto
-                  `AppE` table
-                  `AppE` queryClauseExp
-                  `AppE` ConE 'S.OnConflictDoRaise
-                  `AppE` returning
+                      "INSERT INTO ... VALUES must specify column names for the Squeal-QQ translation."
+            -- Case 2: INSERT ... SELECT ... (a general SELECT statement)
+            _ ->
+              -- selectStmt is not a ValuesSimpleSelect (i.e., it's a general query)
+              case maybeInsertColumnList of
+                Just _ ->
+                  fail
+                    "INSERT INTO table (columns) SELECT ... is not yet supported by Squeal-QQ. Please use INSERT INTO table SELECT ... and ensure your SELECT statement provides all columns for the table, matching the table's column order and types."
+                Nothing -> do
+                  squealQueryExp <- toSquealQuery cteNames Nothing selectStmt -- from Squeal.QuasiQuotes.Query
+                  pure (ConE 'S.Subquery `AppE` squealQueryExp)
         PGT_AST.DefaultValuesInsertRest ->
           fail "INSERT INTO ... DEFAULT VALUES is not yet supported by Squeal-QQ."
 
+    let
+      renderReturning maybeReturning = case maybeReturning of
+        Nothing -> pure $ ConE 'S.Returning_ `AppE` ConE 'S.Nil
+        Just (NE.toList -> [PGT_AST.AsteriskTargetEl]) ->
+          pure $ ConE 'S.Returning `AppE` ConE 'S.Star
+        Just targetList -> do
+          returningProj <- renderTargetList cteNames (NE.toList targetList)
+          pure $ ConE 'S.Returning `AppE` (ConE 'S.List `AppE` returningProj)
+    insertBody <-
+      case (maybeOnConflict, maybeReturningClause) of
+        (Nothing, Nothing) ->
+          pure $ VarE 'S.insertInto_ `AppE` table `AppE` queryClauseExp
+        (Just onConflict, returning) -> do
+          onConflictExp <- renderOnConflict cteNames onConflict
+          returningExp <- renderReturning returning
+          pure $ VarE 'S.insertInto `AppE` table `AppE` queryClauseExp `AppE` onConflictExp `AppE` returningExp
+        (Nothing, returning) -> do
+          let onConflictExp = ConE 'S.OnConflictDoRaise
+          returningExp <- renderReturning returning
+          pure $ VarE 'S.insertInto `AppE` table `AppE` queryClauseExp `AppE` onConflictExp `AppE` returningExp
+
     let
       finalExp = case renderedWithClause of
         Nothing -> insertBody
@@ -253,3 +250,79 @@ renderPGTValueRow cteNames colItems exprs
             `AppE` LabelE colNameStr
 
 
+--
+-- ON CONFLICT support note
+-- -------------------------
+-- As of the pinned upstream `squeal-postgresql` (0.9.2.0), the
+-- `ConflictTarget` GADT only exposes `OnConstraint` and there is no
+-- clause form for a targetless `ON CONFLICT DO NOTHING`.
+--
+-- Consequently, this quasiquoter intentionally fails fast when the parsed
+-- `PostgresqlSyntax` AST requests either:
+--   * a targetless ON CONFLICT (i.e. DO NOTHING without a target), or
+--   * a column-list conflict target (i.e. ON CONFLICT (col[, ...]) ...).
+--
+-- When upstream adds the necessary constructors (e.g. an `OnColumns`
+-- conflict target and a clause for targetless `DO NOTHING`), this is the
+-- place to wire them through. See `plan.md` for the backlog item.
+renderOnConflict :: [Text.Text] -> PGT_AST.OnConflict -> Q Exp
+renderOnConflict cteNames (PGT_AST.OnConflict maybeConfExpr onConflictDo) = do
+  conflictActionExp <- renderOnConflictDo cteNames onConflictDo
+  case maybeConfExpr of
+    Nothing ->
+      fail
+        (  "ON CONFLICT without a conflict target (i.e. targetless DO NOTHING) "
+        <> "cannot be represented with squeal-postgresql 0.9.2.0. "
+        <> "Only ON CONFLICT ON CONSTRAINT ... is currently available. "
+        <> "See plan.md for the proposed upstream changes and status."
+        )
+    Just confExpr -> do
+      conflictTargetExp <- renderConfExpr confExpr
+      pure $ ConE 'S.OnConflict `AppE` conflictTargetExp `AppE` conflictActionExp
+
+-- See the ON CONFLICT support note above.
+renderConfExpr :: PGT_AST.ConfExpr -> Q Exp
+renderConfExpr = \case
+  PGT_AST.ConstraintConfExpr name ->
+    pure $ ConE 'S.OnConstraint `AppE` LabelE (Text.unpack (getIdentText name))
+  PGT_AST.WhereConfExpr _ _ ->
+    fail
+      (  "ON CONFLICT (columns ...) is not supported by squeal-postgresql 0.9.2.0. "
+      <> "Only ON CONFLICT ON CONSTRAINT ... is currently available. "
+      <> "See plan.md for the proposed upstream changes and status."
+      )
+
+renderOnConflictDo :: [Text.Text] -> PGT_AST.OnConflictDo -> Q Exp
+renderOnConflictDo cteNames = \case
+  PGT_AST.NothingOnConflictDo -> pure $ ConE 'S.DoNothing
+  PGT_AST.UpdateOnConflictDo setClauseList maybeWhereClause -> do
+    setClauseListExp <- renderPGTSetClauseList' cteNames setClauseList
+    whereClauseExp <- case maybeWhereClause of
+      Nothing -> pure $ ListE []
+      Just whereClause -> ListE . (:[]) <$> renderPGTAExpr cteNames whereClause
+    pure $ ConE 'S.DoUpdate `AppE` setClauseListExp `AppE` whereClauseExp
+
+renderPGTSetClauseList' :: [Text.Text] -> PGT_AST.SetClauseList -> Q Exp
+renderPGTSetClauseList' cteNames setClauses = do
+  renderedItems <- mapM (renderPGTSetClause' cteNames) (NE.toList setClauses)
+  pure $
+    foldr
+      (\item acc -> ConE '(S.:*) `AppE` item `AppE` acc)
+      (ConE 'S.Nil)
+      renderedItems
+
+renderPGTSetClause' :: [Text.Text] -> PGT_AST.SetClause -> Q Exp
+renderPGTSetClause' cteNames = \case
+  PGT_AST.TargetSetClause (PGT_AST.SetTarget colId maybeIndirection) aExpr -> do
+    when (isJust maybeIndirection) $
+      fail "UPDATE SET with indirection (e.g., array access) is not supported."
+    let
+      colNameStr = Text.unpack (getIdentText colId)
+    renderedExpr <- renderPGTAExpr cteNames aExpr
+    pure $
+      VarE 'S.as
+        `AppE` (ConE 'S.Set `AppE` renderedExpr)
+        `AppE` LabelE colNameStr
+  PGT_AST.TargetListSetClause _ _ ->
+    fail
+      "UPDATE with multiple SET targets (e.g. (col1, col2) = (val1, val2)) is not yet supported."
