@@ -16,10 +16,13 @@ module Squeal.QuasiQuotes.Query (
   getIdentText,
 ) where
 
-import Control.Applicative (Alternative((<|>)))
-import Control.Monad (unless, when)
-import Data.Foldable (Foldable(elem, foldl'), foldlM)
+import Control.Monad (when, zipWithM)
+import Data.Either (partitionEithers)
+import Data.Foldable (Foldable(elem, foldl', foldr, length, null), any, foldlM, mapM_)
+import Data.Function (on)
+import Data.List (groupBy, partition, sortBy)
 import Data.Maybe (fromMaybe, isJust, isNothing)
+import Data.Ord (comparing)
 import Data.String (IsString(fromString))
 import Language.Haskell.TH.Syntax
   ( Exp(AppE, AppTypeE, ConE, InfixE, LabelE, ListE, LitE, TupE, VarE)
@@ -27,10 +30,10 @@ import Language.Haskell.TH.Syntax
   )
 import Prelude
   ( Applicative(pure), Bool(False, True), Either(Left, Right), Eq((==))
-  , Foldable(foldr, length, null), Functor(fmap), Maybe(Just, Nothing)
-  , MonadFail(fail), Num((*), (+), (-), fromInteger), Ord((<), (>=))
-  , Semigroup((<>)), Show(show), Traversable(mapM), ($), (&&), (.), (<$>), (||)
-  , Int, Integer, any, either, error, fromIntegral, id, otherwise, zip
+  , Functor(fmap), Maybe(Just, Nothing), MonadFail(fail)
+  , Num((*), (+), (-), fromInteger), Ord((<), (>=), compare), Semigroup((<>))
+  , Show(show), Traversable(mapM), ($), (&&), (++), (.), (<$>), (||), Int
+  , Integer, either, error, fromIntegral, id, maybe, otherwise, uncurry, zip
   )
 import qualified Data.ByteString.Char8 as BS8
 import qualified Data.List.NonEmpty as NE
@@ -39,6 +42,49 @@ import qualified PostgresqlSyntax.Ast as PGT_AST
 import qualified Squeal.PostgreSQL as S
 
 
+-- | Intermediate representation for a window function call in a SELECT list.
+data WindowFuncInfo = WindowFuncInfo
+  { wfiTargetEl :: PGT_AST.TargetEl
+  , wfiFuncApp :: PGT_AST.FuncApplication
+  , wfiOverClause :: PGT_AST.OverClause
+  }
+  deriving stock (Eq, Show)
+
+
+-- | A wrapper for `PGT_AST.OverClause` to provide an `Ord` instance for sorting and grouping.
+newtype OrdOverClause = OrdOverClause PGT_AST.OverClause
+  deriving stock (Eq, Show)
+
+
+-- Manual Ord instance based on the rendered string representation for simplicity.
+instance Ord OrdOverClause where
+  compare (OrdOverClause c1) (OrdOverClause c2) =
+    compare (show c1) (show c2)
+
+
+-- | `WindowFuncInfo` using `OrdOverClause`.
+data WindowFuncInfo_ = WindowFuncInfo_
+  { wfiTargetEl_ :: PGT_AST.TargetEl
+  , wfiOverClause_ :: OrdOverClause
+  }
+
+
+-- | Classifies a `PGT_AST.TargetEl` as either a normal expression or a window function call.
+isWindowTarget :: PGT_AST.TargetEl -> Either PGT_AST.TargetEl WindowFuncInfo_
+isWindowTarget el = case el of
+    PGT_AST.AliasedExprTargetEl expr _ -> go expr
+    PGT_AST.ImplicitlyAliasedExprTargetEl expr _ -> go expr
+    PGT_AST.ExprTargetEl expr -> go expr
+    _ -> Left el
+  where
+    go
+      ( PGT_AST.CExprAExpr
+          (PGT_AST.FuncCExpr (PGT_AST.ApplicationFuncExpr _app _ _ (Just over)))
+        ) =
+        Right $ WindowFuncInfo_ el (OrdOverClause over)
+    go _ = Left el
+
+
 toSquealQuery
   :: [Text.Text]
   -> Maybe (NE.NonEmpty PGT_AST.Ident)
@@ -61,6 +107,23 @@ toSquealSelectWithParens cteNames maybeColAliases = \case
     toSquealSelectWithParens cteNames maybeColAliases swp
 
 
+toSquealSelectClause
+  :: [Text.Text]
+  -> Maybe (NE.NonEmpty PGT_AST.Ident)
+  -> PGT_AST.SelectClause
+  -> Q Exp
+toSquealSelectClause cteNames maybeColAliases = \case
+  Left simpleSelect ->
+    toSquealSimpleSelect
+      cteNames
+      maybeColAliases
+      simpleSelect
+      Nothing
+      Nothing
+      Nothing
+  Right selectWithParens -> toSquealSelectWithParens cteNames maybeColAliases selectWithParens
+
+
 toSquealSelectNoParens
   :: [Text.Text]
   -> Maybe (NE.NonEmpty PGT_AST.Ident)
@@ -76,12 +139,10 @@ toSquealSelectNoParens
       maybeSelectLimit
       maybeForLockingClause
     ) = do
-    (cteNames, renderedWithClause) <-
+    (cteNames, withApp) <-
       case maybeWithClause of
-        Nothing -> pure (initialCteNames, Nothing)
-        Just withClause -> do
-          (names, exp) <- renderPGTWithClause initialCteNames withClause
-          pure (names, Just exp)
+        Nothing -> pure (initialCteNames, id)
+        Just withClause -> renderPGTWithClause initialCteNames withClause
 
     squealQueryBody <-
       case selectClause of
@@ -95,32 +156,42 @@ toSquealSelectNoParens
             maybeForLockingClause
         Right selectWithParens' -> toSquealSelectWithParens cteNames maybeColAliases selectWithParens'
 
-    case renderedWithClause of
-      Nothing -> pure squealQueryBody
-      Just withExp -> pure $ VarE 'S.with `AppE` withExp `AppE` squealQueryBody
+    pure $ withApp squealQueryBody
 
 
-renderPGTWithClause :: [Text.Text] -> PGT_AST.WithClause -> Q ([Text.Text], Exp)
-renderPGTWithClause initialCteNames (PGT_AST.WithClause recursive ctes) = do
-    when recursive $ fail "Recursive WITH clauses are not supported yet."
-    let
-      cteList = NE.toList ctes
-    (finalCteNames, renderedCtes) <-
-      foldlM
-        ( \(names, exps) cte -> do
-            (name, exp) <- renderCte names cte
-            pure (names <> [name], exps <> [exp])
-        )
-        (initialCteNames, [])
-        cteList
+renderPGTWithClause
+  :: [Text.Text] -> PGT_AST.WithClause -> Q ([Text.Text], Exp -> Exp)
+renderPGTWithClause initialCteNames (PGT_AST.WithClause recursive ctes) =
+    if recursive
+      then do
+        case NE.toList ctes of
+          [cte] -> do
+            (cteName, aliasedCteQueryExp) <- renderRecursiveCte initialCteNames cte
+            let
+              withApp body = VarE 'S.withRecursive `AppE` aliasedCteQueryExp `AppE` body
+            pure (initialCteNames <> [cteName], withApp)
+          _ -> fail "Squeal-QQ currently only supports WITH RECURSIVE with a single CTE."
+      else do
+        let
+          cteList = NE.toList ctes
+        (finalCteNames, renderedCtes) <-
+          foldlM
+            ( \(names, exps) cte -> do
+                (name, exp) <- renderCte names cte
+                pure (names <> [name], exps <> [exp])
+            )
+            (initialCteNames, [])
+            cteList
 
-    let
-      withExp =
-        foldr
-          (\cte acc -> ConE '(S.:>>) `AppE` cte `AppE` acc)
-          (ConE 'S.Done)
-          renderedCtes
-    pure (finalCteNames, withExp)
+        let
+          withExp =
+            foldr
+              (\cte acc -> ConE '(S.:>>) `AppE` cte `AppE` acc)
+              (ConE 'S.Done)
+              renderedCtes
+        let
+          withApp body = VarE 'S.with `AppE` withExp `AppE` body
+        pure (finalCteNames, withApp)
   where
     renderCte :: [Text.Text] -> PGT_AST.CommonTableExpr -> Q (Text.Text, Exp)
     renderCte existingCteNames (PGT_AST.CommonTableExpr ident maybeColNames maybeMaterialized stmt) = do
@@ -137,6 +208,26 @@ renderPGTWithClause initialCteNames (PGT_AST.WithClause recursive ctes) = do
       pure
         (cteName, VarE 'S.as `AppE` cteQueryExp `AppE` LabelE (Text.unpack cteName))
 
+    renderRecursiveCte
+      :: [Text.Text] -> PGT_AST.CommonTableExpr -> Q (Text.Text, Exp)
+    renderRecursiveCte existingCteNames (PGT_AST.CommonTableExpr ident maybeColNames maybeMaterialized stmt) = do
+      when (isJust maybeColNames) $
+        fail "Column name lists in CTEs are not supported yet."
+      when (isJust maybeMaterialized) $
+        fail "MATERIALIZED/NOT MATERIALIZED for CTEs is not supported yet."
+      let
+        cteName = getIdentText ident
+      -- For a recursive CTE, its own name must be in scope for the query inside it.
+      let
+        ctesInScope = existingCteNames <> [cteName]
+      cteQueryExp <-
+        case stmt of
+          PGT_AST.SelectPreparableStmt selectStmt -> toSquealQuery ctesInScope Nothing selectStmt
+          _ -> fail "Only SELECT statements are supported in CTEs."
+      let
+        aliasedQuery = VarE 'S.as `AppE` cteQueryExp `AppE` LabelE (Text.unpack cteName)
+      pure (cteName, aliasedQuery)
+
 
 toSquealSimpleSelect
   :: [Text.Text]
@@ -148,17 +239,39 @@ toSquealSimpleSelect
   -> Q Exp
 toSquealSimpleSelect cteNames maybeColAliases simpleSelect maybeSortClause maybeSelectLimit maybeForLockingClause =
   case simpleSelect of
+    PGT_AST.BinSimpleSelect op left allOrDistinct right -> do
+      when
+        ( isJust maybeSortClause
+            || isJust maybeSelectLimit
+            || isJust maybeForLockingClause
+        )
+        $ fail
+          "ORDER BY, LIMIT, OFFSET, and FOR clauses are not supported on the immediate operands of a set operation. You can use parentheses to specify precedence."
+
+      leftQuery <- toSquealSelectClause cteNames maybeColAliases left
+      rightQuery <- toSquealSelectClause cteNames maybeColAliases right
+
+      let
+        squealOp = case (op, allOrDistinct) of
+          (PGT_AST.UnionSelectBinOp, Just False) -> VarE 'S.unionAll
+          (PGT_AST.UnionSelectBinOp, _) -> VarE 'S.union
+          (PGT_AST.IntersectSelectBinOp, Just False) -> VarE 'S.intersectAll
+          (PGT_AST.IntersectSelectBinOp, _) -> VarE 'S.intersect
+          (PGT_AST.ExceptSelectBinOp, Just False) -> VarE 'S.exceptAll
+          (PGT_AST.ExceptSelectBinOp, _) -> VarE 'S.except
+
+      pure $ squealOp `AppE` leftQuery `AppE` rightQuery
     PGT_AST.ValuesSimpleSelect valuesClause -> do
-      unless
-        ( isNothing maybeSortClause
-            && isNothing maybeSelectLimit
-            && isNothing maybeForLockingClause
+      when
+        ( isJust maybeSortClause
+            || isJust maybeSelectLimit
+            || isJust maybeForLockingClause
         )
         $ fail
         $ "ORDER BY / OFFSET / LIMIT / FOR UPDATE etc. not supported with VALUES clause "
           <> "in this translation yet."
-      renderedValues <- renderValuesClauseToNP cteNames maybeColAliases valuesClause
-      pure $ VarE 'S.values_ `AppE` renderedValues
+      valuesExp <- renderValuesClauseToNP cteNames maybeColAliases valuesClause
+      pure valuesExp
     PGT_AST.NormalSimpleSelect
       maybeTargeting
       maybeIntoClause
@@ -213,9 +326,7 @@ toSquealSimpleSelect cteNames maybeColAliases simpleSelect maybeSortClause maybe
                 when (isJust maybeIntoClause) $
                   fail "INTO clause is not yet supported in this translation."
                 when (isJust maybeWindowClause) $
-                  fail $
-                    "WINDOW clause is not yet supported in this translation "
-                      <> "for NormalSimpleSelect with FROM."
+                  fail "WINDOW clause is not yet supported."
 
                 renderedFromClauseExp <- renderPGTTableRef cteNames fromClause
                 let
@@ -318,45 +429,63 @@ renderValuesClauseToNP
   -> PGT_AST.ValuesClause
   -> Q Exp
 renderValuesClauseToNP cteNames maybeColAliases (firstRowExps NE.:| restRowExps) = do
-  unless (null restRowExps) $
-    fail $
-      "Multi-row VALUES clause requires S.values, this translation "
-        <> "currently supports single row S.values_."
-  convertRowToNP firstRowExps
- where
-  colAliasTexts = fmap (fmap getIdentText . NE.toList) maybeColAliases
-
-  convertRowToNP :: NE.NonEmpty PGT_AST.AExpr -> Q Exp
-  convertRowToNP exprs = do
     let
-      exprList = NE.toList exprs
+      firstRowList = NE.toList firstRowExps
+      restRowsLists = fmap NE.toList restRowExps
+
+    -- Determine column aliases based on provided aliases or synthesize.
     aliasTexts <-
-      case colAliasTexts of
+      case fmap (fmap getIdentText . NE.toList) maybeColAliases of
         Just aliases ->
-          if length aliases == length exprList
+          if length aliases == length firstRowList
             then pure aliases
             else
               fail
-                "Number of column aliases in CTE does not match number of columns in VALUES clause."
-        Nothing -> pure $ fmap (Text.pack . ("_column" <>) . show) [1 :: Int ..]
-    go (zip exprList aliasTexts)
-   where
-    go :: [(PGT_AST.AExpr, Text.Text)] -> Q Exp
-    go [] = pure $ ConE 'S.Nil
-    go ((expr, aliasText) : fs) = do
-      renderedExpr <- renderPGTAExpr cteNames expr
-      let
-        aliasedExp = VarE 'S.as `AppE` renderedExpr `AppE` LabelE (Text.unpack aliasText)
-      restExp <- go fs
-      pure $ ConE '(S.:*) `AppE` aliasedExp `AppE` restExp
+                "Number of column aliases does not match number of columns in VALUES clause."
+        Nothing ->
+          pure $ fmap (Text.pack . ("_column" <>) . show) [1 :: Int ..]
+
+    -- Validate all rows have consistent arity.
+    let expectedLen = length firstRowList
+    let checkLen xs = if length xs == expectedLen
+          then pure ()
+          else fail "Mismatched number of columns across VALUES rows."
+    mapM_ checkLen restRowsLists
+
+    -- Helper to convert a row to an NP using the established aliases.
+    let convertRow :: [PGT_AST.AExpr] -> Q Exp
+        convertRow exprs = go (zip exprs aliasTexts)
+          where
+            go :: [(PGT_AST.AExpr, Text.Text)] -> Q Exp
+            go [] = pure $ ConE 'S.Nil
+            go ((expr, aliasText) : fs) = do
+              renderedExpr <- renderPGTAExpr cteNames expr
+              let aliasedExp = VarE 'S.as `AppE` renderedExpr `AppE` LabelE (Text.unpack aliasText)
+              restExp <- go fs
+              pure $ ConE '(S.:*) `AppE` aliasedExp `AppE` restExp
+
+    firstRowNP <- convertRow firstRowList
+    case restRowsLists of
+      [] -> pure $ VarE 'S.values_ `AppE` firstRowNP
+      more -> do
+        moreNPs <- mapM convertRow more
+        pure $ VarE 'S.values `AppE` firstRowNP `AppE` ListE moreNPs
 
 
 renderPGTForLockingClauseItems :: PGT_AST.ForLockingClause -> Q [Exp]
 renderPGTForLockingClauseItems = \case
+  -- PostgreSQL's `FOR READ ONLY` does not acquire row-level locks and is
+  -- effectively equivalent to the absence of a row-locking clause for
+  -- SELECT. Squeal does not expose a row-level "read only" lock; instead,
+  -- read-only behavior is an AccessMode on the transaction
+  -- (TransactionMode { accessMode = ReadOnly }).
+  --
+  -- We therefore accept and ignore `FOR READ ONLY` here so that the quoted
+  -- SQL compiles and renders identically to the same query without the
+  -- clause. To enforce read-only semantics, execute the statement inside a
+  -- transaction whose AccessMode is ReadOnly.
   PGT_AST.ReadOnlyForLockingClause ->
-    fail $
-      "FOR READ ONLY is not supported as a row-level locking "
-        <> "clause by Squeal-QQ."
+    pure []
   PGT_AST.ItemsForLockingClause itemsNe ->
     mapM renderPGTForLockingItem (NE.toList itemsNe)
 
@@ -439,6 +568,7 @@ renderPGTGroupByClauseElements cteNames = \case
         (ConE 'S.Nil)
         renderedExprs
 
+
 renderPGTGroupByItem :: [Text.Text] -> PGT_AST.GroupByItem -> Q Exp
 renderPGTGroupByItem cteNames = \case
   PGT_AST.ExprGroupByItem scalarExpr -> renderPGTAExpr cteNames scalarExpr
@@ -448,7 +578,8 @@ renderPGTGroupByItem cteNames = \case
       "Unsupported grouping expression: " <> show unsupportedGroup
 
 
-processSelectLimit :: [Text.Text] -> Exp -> Maybe PGT_AST.SelectLimit -> Q (Exp, Maybe Exp)
+processSelectLimit
+  :: [Text.Text] -> Exp -> Maybe PGT_AST.SelectLimit -> Q (Exp, Maybe Exp)
 processSelectLimit _cteNames tableExpr Nothing = pure (tableExpr, Nothing)
 processSelectLimit cteNames tableExpr (Just selectLimit) = do
   let
@@ -622,16 +753,16 @@ renderPGTTargetingForValues cteNames = \case
 
 renderPGTOnExpressionsClause :: [Text.Text] -> [PGT_AST.AExpr] -> Q Exp
 renderPGTOnExpressionsClause cteNames exprs = do
-  renderedSortExps <- mapM renderToSortExpr exprs
-  pure $ ListE renderedSortExps
- where
-  renderToSortExpr :: PGT_AST.AExpr -> Q Exp
-  renderToSortExpr astExpr = do
-    squealExpr <- renderPGTAExpr cteNames astExpr
-    -- For DISTINCT ON, the direction (ASC/DESC) and NULLS order
-    -- are typically specified in the ORDER BY clause.
-    -- Here, we default to ASC for the SortExpression constructor.
-    pure $ ConE 'S.Asc `AppE` squealExpr
+    renderedSortExps <- mapM renderToSortExpr exprs
+    pure $ ListE renderedSortExps
+  where
+    renderToSortExpr :: PGT_AST.AExpr -> Q Exp
+    renderToSortExpr astExpr = do
+      squealExpr <- renderPGTAExpr cteNames astExpr
+      -- For DISTINCT ON, the direction (ASC/DESC) and NULLS order
+      -- are typically specified in the ORDER BY clause.
+      -- Here, we default to ASC for the SortExpression constructor.
+      pure $ ConE 'S.Asc `AppE` squealExpr
 
 
 renderPGTSortClause :: [Text.Text] -> PGT_AST.SortClause -> Q Exp
@@ -675,6 +806,21 @@ renderSingleTableRef cteNames = \case
   PGT_AST.RelationExprTableRef relationExpr maybeAliasClause sampleClause -> do
     when (isJust sampleClause) $ fail "TABLESAMPLE clause is not supported yet."
     renderPGTRelationExprTableRef cteNames relationExpr maybeAliasClause
+  -- Support subqueries (including VALUES ...) in FROM with optional column aliases
+  PGT_AST.SelectTableRef isLateral selectWithParens maybeAliasClause -> do
+    when isLateral $ fail "LATERAL subqueries are not supported yet."
+    -- Alias is required for subqueries in FROM in PostgreSQL; enforce here
+    (aliasStr, maybeColAliases) <-
+      case maybeAliasClause of
+        Just (PGT_AST.AliasClause _ aliasIdent maybeCols) -> do
+          let aliasTxt = Text.unpack (getIdentText aliasIdent)
+          pure (aliasTxt, maybeCols)
+        Nothing -> fail "Subquery in FROM requires an alias."
+    -- Propagate any provided column aliases down into the subquery translation
+    subqueryExp <- toSquealSelectWithParens cteNames maybeColAliases selectWithParens
+    -- Treat the subquery as a derived table in the FROM clause
+    -- Using S.subquery (as subquery alias)
+    pure $ VarE 'S.subquery `AppE` (VarE 'S.as `AppE` subqueryExp `AppE` LabelE aliasStr)
   PGT_AST.JoinTableRef joinedTable maybeAliasClause ->
     -- If `maybeAliasClause` is Just, it means `(JOIN_TABLE) AS alias`.
     -- Squeal's direct join combinators don't alias the *result* of the join.
@@ -830,64 +976,122 @@ renderPGTTargeting cteNames = \case
     pure (selListExp, fmap NE.toList maybeOnExprs)
 
 
-renderPGTTargetEl :: [Text.Text] -> PGT_AST.TargetEl -> Maybe PGT_AST.Ident -> Int -> Q Exp
-renderPGTTargetEl cteNames targetEl mOuterAlias idx =
+renderPGTTargetEl :: [Text.Text] -> PGT_AST.TargetEl -> Int -> Q Exp
+renderPGTTargetEl cteNames targetEl idx =
   let
     (exprAST, mInternalAlias) = case targetEl of
       PGT_AST.AliasedExprTargetEl e an -> (e, Just an)
       PGT_AST.ImplicitlyAliasedExprTargetEl e an -> (e, Just an)
       PGT_AST.ExprTargetEl e -> (e, Nothing)
-      PGT_AST.AsteriskTargetEl ->
-        ( PGT_AST.CExprAExpr
-            ( PGT_AST.AexprConstCExpr
-                PGT_AST.NullAexprConst
-            )
-        , Nothing -- Placeholder for Star, should be S.Star
-        )
-    finalAliasName = mOuterAlias <|> mInternalAlias
+      _ -> error "renderPGTTargetEl called with non-expression TargetEl"
   in
-    case targetEl of
-      PGT_AST.AsteriskTargetEl -> pure $ ConE 'S.Star
-      _ -> do
-        renderedScalarExp <- renderPGTAExpr cteNames exprAST
-        case exprAST of
-          PGT_AST.CExprAExpr (PGT_AST.ColumnrefCExpr _)
-            | Nothing <- finalAliasName ->
-                pure renderedScalarExp
-          _ -> do
-            let
-              aliasLabelStr =
-                case finalAliasName of
-                  Just ident -> Text.unpack $ getIdentText ident
-                  Nothing -> "_col" <> show idx
-            pure $
-              VarE 'S.as
-                `AppE` renderedScalarExp
-                `AppE` LabelE aliasLabelStr
+    do
+      renderedScalarExp <- renderPGTAExpr cteNames exprAST
+      case exprAST of
+        PGT_AST.CExprAExpr (PGT_AST.ColumnrefCExpr _)
+          | Nothing <- mInternalAlias ->
+              pure renderedScalarExp
+        _ -> do
+          let
+            aliasLabelStr =
+              case mInternalAlias of
+                Just ident -> Text.unpack $ getIdentText ident
+                Nothing -> "_col" <> show idx
+          pure $
+            VarE 'S.as
+              `AppE` renderedScalarExp
+              `AppE` LabelE aliasLabelStr
 
 
 renderPGTTargetList :: [Text.Text] -> PGT_AST.TargetList -> Q Exp
-renderPGTTargetList cteNames (item NE.:| items) = go (item : items) 1
+renderPGTTargetList cteNames (item NE.:| items) = do
+  let
+    allItems = item : items
+    (normalTargets, windowTargets) = partitionEithers (isWindowTarget <$> allItems)
+
+  -- Group window functions by their OVER clause
+  let
+    sortedWindowTargets = sortBy (comparing wfiOverClause_) windowTargets
+    groupedWindowTargets = groupBy ((==) `on` wfiOverClause_) sortedWindowTargets
+
+  -- Render normal targets
+  renderedNormalSelections <-
+    if null normalTargets
+      then pure []
+      else (: []) <$> renderNormalTargetList cteNames normalTargets
+
+  -- Render window target groups
+  (_, renderedWindowSelections) <-
+    foldlM
+      ( \(idx, acc) grp -> do
+          (newIdx, renderedGrp) <- renderWindowGroup cteNames idx grp
+          pure (newIdx, acc ++ [renderedGrp])
+      )
+      (1, [])
+      groupedWindowTargets
+
+  -- Combine all selections
+  let
+    allSelections = renderedNormalSelections ++ renderedWindowSelections
+  case allSelections of
+    [] -> fail "Empty selection list"
+    [sel] -> pure sel
+    (sel : sels) -> pure $ foldl' (\acc s -> ConE 'S.Also `AppE` s `AppE` acc) sel sels
+
+
+renderNormalTargetList :: [Text.Text] -> [PGT_AST.TargetEl] -> Q Exp
+renderNormalTargetList cteNames targets = do
+    let
+      isAsterisk :: PGT_AST.TargetEl -> Bool
+      isAsterisk PGT_AST.AsteriskTargetEl = True
+      isAsterisk _ = False
+
+      isDotStar :: PGT_AST.TargetEl -> Bool
+      isDotStar
+        ( PGT_AST.ExprTargetEl
+            ( PGT_AST.CExprAExpr
+                (PGT_AST.ColumnrefCExpr (PGT_AST.Columnref _ (Just indirection)))
+              )
+          ) =
+          any isAllIndirectionEl (NE.toList indirection)
+      isDotStar _ = False
+
+      isAllIndirectionEl :: PGT_AST.IndirectionEl -> Bool
+      isAllIndirectionEl PGT_AST.AllIndirectionEl = True
+      isAllIndirectionEl _ = False
+
+    let
+      (stars, notStars) = partition isAsterisk targets
+      (dotStars, normalExprs) = partition isDotStar notStars
+
+    renderedStar <-
+      case stars of
+        [] -> pure Nothing
+        [_] -> pure $ Just (ConE 'S.Star)
+        _ -> fail "Multiple `*` in SELECT list is not supported."
+
+    renderedDotStars <- mapM renderPGTTargetElDotStar dotStars
+
+    renderedNormalsExp <-
+      if null normalExprs
+        then pure Nothing
+        else do
+          renderedEls <- zipWithM (renderPGTTargetEl cteNames) normalExprs [1 ..]
+          let
+            npList = foldr (\h t -> ConE '(S.:*) `AppE` h `AppE` t) (ConE 'S.Nil) renderedEls
+          pure $ Just (ConE 'S.List `AppE` npList)
+
+    let
+      allParts =
+        maybe [] pure renderedStar
+          ++ renderedDotStars
+          ++ maybe [] pure renderedNormalsExp
+
+    case allParts of
+      [] -> fail "Empty normal selection list"
+      [sel] -> pure sel
+      (sel : sels) -> pure $ foldl' (\acc s -> ConE 'S.Also `AppE` s `AppE` acc) sel sels
   where
-    isAsterisk :: PGT_AST.TargetEl -> Bool
-    isAsterisk PGT_AST.AsteriskTargetEl = True
-    isAsterisk _ = False
-
-    isDotStar :: PGT_AST.TargetEl -> Bool
-    isDotStar
-      ( PGT_AST.ExprTargetEl
-          ( PGT_AST.CExprAExpr
-              (PGT_AST.ColumnrefCExpr (PGT_AST.Columnref _ (Just indirection)))
-            )
-        ) =
-        any isAllIndirectionEl (NE.toList indirection)
-    isDotStar _ = False
-
-    isAllIndirectionEl :: PGT_AST.IndirectionEl -> Bool
-    isAllIndirectionEl PGT_AST.AllIndirectionEl = True
-    isAllIndirectionEl _ = False
-
-    renderPGTTargetElDotStar :: PGT_AST.TargetEl -> Q Exp
     renderPGTTargetElDotStar
       ( PGT_AST.ExprTargetEl
           ( PGT_AST.CExprAExpr
@@ -905,19 +1109,131 @@ renderPGTTargetList cteNames (item NE.:| items) = go (item : items) 1
     renderPGTTargetElDotStar _ =
       fail "renderPGTTargetElDotStar called with unexpected TargetEl"
 
-    go :: [PGT_AST.TargetEl] -> Int -> Q Exp
-    go [] _ = fail "Empty selection list items in go."
-    go [el] currentIdx = renderOne el currentIdx
-    go (el : more) currentIdx = do
-      renderedEl <- renderOne el currentIdx
-      restRendered <- go more (currentIdx + 1)
-      pure $ ConE 'S.Also `AppE` restRendered `AppE` renderedEl
 
-    renderOne :: PGT_AST.TargetEl -> Int -> Q Exp
-    renderOne el idx
-      | isAsterisk el = pure $ ConE 'S.Star
-      | isDotStar el = renderPGTTargetElDotStar el
-      | otherwise = renderPGTTargetEl cteNames el Nothing idx
+renderWindowGroup :: [Text.Text] -> Int -> [WindowFuncInfo_] -> Q (Int, Exp)
+renderWindowGroup cteNames startIdx = \case
+  [] -> fail "renderWindowGroup: received an empty group, this should not happen."
+  group@(head_info : _) -> do
+    let
+      OrdOverClause overClause = wfiOverClause_ head_info
+    windowDefExp <-
+      case overClause of
+        PGT_AST.ColIdOverClause _ -> fail "WINDOW clause with named windows is not supported yet."
+        PGT_AST.WindowOverClause spec -> renderPGTWindowSpecification cteNames spec
+
+    (newIdx, windowFuncsNP) <-
+      renderWindowFuncsNP cteNames startIdx (wfiTargetEl_ <$> group)
+
+    pure $ (newIdx, ConE 'S.Over `AppE` windowFuncsNP `AppE` windowDefExp)
+
+
+renderPGTWindowSpecification
+  :: [Text.Text] -> PGT_AST.WindowSpecification -> Q Exp
+renderPGTWindowSpecification cteNames (PGT_AST.WindowSpecification mExisting mPartition mSort mFrame) = do
+  when (isJust mExisting) $ fail "Existing window names are not supported yet."
+  when (isJust mFrame) $
+    fail "Frame clauses (ROWS/RANGE/GROUPS) are not supported yet."
+
+  partitionByExp <-
+    case mPartition of
+      Nothing -> pure $ VarE 'S.partitionBy `AppE` ConE 'S.Nil
+      Just partitionExps -> do
+        renderedExps <- mapM (renderPGTAExpr cteNames) partitionExps
+        let
+          np = foldr (\h t -> ConE '(S.:*) `AppE` h `AppE` t) (ConE 'S.Nil) renderedExps
+        pure $ VarE 'S.partitionBy `AppE` np
+
+  case mSort of
+    Nothing -> pure partitionByExp
+    Just sortClause -> do
+      renderedSC <- renderPGTSortClause cteNames sortClause
+      pure $
+        InfixE
+          (Just partitionByExp)
+          (VarE '(S.&))
+          (Just (VarE 'S.orderBy `AppE` renderedSC))
+
+
+renderWindowFuncsNP :: [Text.Text] -> Int -> [PGT_AST.TargetEl] -> Q (Int, Exp)
+renderWindowFuncsNP cteNames startIdx targets = do
+  let
+    indexedTargets = zip targets [startIdx ..]
+  renderedFuncs <-
+    mapM (uncurry (renderWindowFuncAsAliasedNP cteNames)) indexedTargets
+  let
+    newIdx = startIdx + length targets
+  pure $
+    ( newIdx
+    , foldr (\h t -> ConE '(S.:*) `AppE` h `AppE` t) (ConE 'S.Nil) renderedFuncs
+    )
+
+
+renderWindowFuncAsAliasedNP :: [Text.Text] -> PGT_AST.TargetEl -> Int -> Q Exp
+renderWindowFuncAsAliasedNP cteNames el idx = do
+  let
+    (funcApp, mAlias) = case el of
+      PGT_AST.AliasedExprTargetEl
+        (PGT_AST.CExprAExpr (PGT_AST.FuncCExpr (PGT_AST.ApplicationFuncExpr app _ _ _)))
+        an -> (app, Just an)
+      PGT_AST.ImplicitlyAliasedExprTargetEl
+        (PGT_AST.CExprAExpr (PGT_AST.FuncCExpr (PGT_AST.ApplicationFuncExpr app _ _ _)))
+        an -> (app, Just an)
+      PGT_AST.ExprTargetEl
+        (PGT_AST.CExprAExpr (PGT_AST.FuncCExpr (PGT_AST.ApplicationFuncExpr app _ _ _))) -> (app, Nothing)
+      _ -> error "renderWindowFuncAsAliasedNP: not a window function"
+
+  let
+    aliasStr = case mAlias of
+      Just ident -> Text.unpack (getIdentText ident)
+      Nothing -> "_window" <> show idx
+
+  renderedFunc <- renderPGTFuncAppAsWindowFunc cteNames funcApp
+
+  pure $ VarE 'S.as `AppE` renderedFunc `AppE` LabelE aliasStr
+
+
+renderPGTFuncAppAsWindowFunc :: [Text.Text] -> PGT_AST.FuncApplication -> Q Exp
+renderPGTFuncAppAsWindowFunc cteNames (PGT_AST.FuncApplication funcName maybeParams) = do
+  (squealFn, isCount) <-
+    case funcName of
+      PGT_AST.TypeFuncName fident -> do
+        let
+          fnNameStr = Text.toLower (getIdentText fident)
+        case fnNameStr of
+          "rank" -> pure (VarE 'S.rank, False)
+          "row_number" -> pure (VarE 'S.rowNumber, False)
+          "dense_rank" -> pure (VarE 'S.denseRank, False)
+          "percent_rank" -> pure (VarE 'S.percentRank, False)
+          "cume_dist" -> pure (VarE 'S.cumeDist, False)
+          "ntile" -> pure (VarE 'S.ntile, False)
+          "lag" -> pure (VarE 'S.lag, False)
+          "lead" -> pure (VarE 'S.lead, False)
+          "first_value" -> pure (VarE 'S.firstValue, False)
+          "last_value" -> pure (VarE 'S.lastValue, False)
+          "nth_value" -> pure (VarE 'S.nthValue, False)
+          "count" -> pure (VarE 'S.count, True)
+          "sum" -> pure (VarE 'S.sum_, False)
+          "avg" -> pure (VarE 'S.avg, False)
+          "min" -> pure (VarE 'S.min_, False)
+          "max" -> pure (VarE 'S.max_, False)
+          _ -> fail $ "Unsupported window function: " <> Text.unpack fnNameStr
+      _ -> fail "Unsupported function name in window function"
+
+  case maybeParams of
+    Nothing -> pure squealFn
+    Just PGT_AST.StarFuncApplicationParams
+      | isCount -> pure $ VarE 'S.countStar
+      | otherwise ->
+          fail "Star argument is only supported for COUNT in window functions."
+    Just (PGT_AST.NormalFuncApplicationParams (Just True) _ _) ->
+      fail "DISTINCT is not supported for window functions."
+    Just (PGT_AST.NormalFuncApplicationParams _ args _) -> do
+      argExps <- mapM (renderPGTFuncArgExpr cteNames) (NE.toList args)
+      let
+        npArgs = foldr (\h t -> ConE '(S.:*) `AppE` h `AppE` t) (ConE 'S.Nil) argExps
+        windowArg = ConE 'S.Windows `AppE` npArgs
+      pure $ squealFn `AppE` windowArg
+    _ -> fail "Unsupported parameters for window function"
 
 
 -- | Defines associativity of an operator.
@@ -1184,7 +1500,8 @@ renderPGTAExpr cteNames astExpr = case fixOperatorPrecedence astExpr of
       PGT_AST.InAExprReversableOp inExpr ->
         case inExpr of
           PGT_AST.ExprListInExpr exprList -> do
-            let opVar' = if not then VarE 'S.notIn else VarE 'S.in_
+            let
+              opVar' = if not then VarE 'S.notIn else VarE 'S.in_
             listExp' <- ListE <$> mapM (renderPGTAExpr cteNames) (NE.toList exprList)
             pure $ opVar' `AppE` renderedExpr' `AppE` listExp'
           PGT_AST.SelectInExpr selectWithParens -> do
@@ -1245,7 +1562,8 @@ renderPGTFuncExpr cteNames = \case
   PGT_AST.ApplicationFuncExpr funcApp maybeWithinGroup maybeFilter maybeOver -> do
     when (isJust maybeWithinGroup) $ fail "WITHIN GROUP clause is not supported."
     when (isJust maybeFilter) $ fail "FILTER clause is not supported."
-    when (isJust maybeOver) $ fail "OVER clause is not supported."
+    when (isJust maybeOver) $
+      fail "OVER clause is only supported at the top level of a SELECT list item."
     renderPGTFuncApplication cteNames funcApp
   PGT_AST.SubexprFuncExpr funcCommonSubexpr -> renderPGTFuncExprCommonSubexpr cteNames funcCommonSubexpr
 
@@ -1258,8 +1576,9 @@ renderPGTFuncApplication cteNames (PGT_AST.FuncApplication funcName maybeParams)
     PGT_AST.TypeFuncName fident ->
       let
         fnNameStr = Text.unpack (getIdentText fident)
+        fnNameStrLower = Text.toLower (getIdentText fident)
       in
-        case Text.toLower (Text.pack fnNameStr) of
+        case fnNameStrLower of
           "inline" ->
             case maybeParams of
               Just (PGT_AST.NormalFuncApplicationParams _ args _) ->
@@ -1286,37 +1605,46 @@ renderPGTFuncApplication cteNames (PGT_AST.FuncApplication funcName maybeParams)
                       pure $ VarE 'S.inlineParam `AppE` VarE varName
                   _ -> fail "inline_param() function expects a single variable argument"
               _ -> fail "inline_param() function expects a single variable argument"
-          otherFnName ->
-            let
-              squealFn :: Q Exp
-              squealFn =
-                case otherFnName of
-                  "coalesce" -> pure $ VarE 'S.coalesce
-                  "lower" -> pure $ VarE 'S.lower
-                  "char_length" -> pure $ VarE 'S.charLength
-                  "character_length" -> pure $ VarE 'S.charLength
-                  "upper" -> pure $ VarE 'S.upper
-                  "count" -> pure $ VarE 'S.count -- Special handling for count(*) might be needed
-                  "now" -> pure $ VarE 'S.now
-                  _ -> fail $ "Unsupported function: " <> fnNameStr
-            in
-              case maybeParams of
-                Nothing -> squealFn -- No-argument function
-                Just params -> case params of
-                  PGT_AST.NormalFuncApplicationParams maybeAllOrDistinct args maybeSortClause -> do
-                    when (isJust maybeAllOrDistinct) $
-                      fail "DISTINCT in function calls is not supported."
-                    when (isJust maybeSortClause) $
-                      fail "ORDER BY in function calls is not supported."
-                    fn <- squealFn
-                    argExps <- mapM (renderPGTFuncArgExpr cteNames) (NE.toList args)
-                    pure $ foldl' AppE fn argExps
-                  PGT_AST.StarFuncApplicationParams ->
-                    -- Specific for count(*)
-                    if fnNameStr == "count"
-                      then pure $ VarE 'S.countStar
-                      else fail "Star argument only supported for COUNT"
-                  _ -> fail $ "Unsupported function parameters structure: " <> show params
+          _ -> do
+            (squealFn, isAggregate) <-
+              case fnNameStrLower of
+                "coalesce" -> pure (VarE 'S.coalesce, False)
+                "lower" -> pure (VarE 'S.lower, False)
+                "char_length" -> pure (VarE 'S.charLength, False)
+                "character_length" -> pure (VarE 'S.charLength, False)
+                "upper" -> pure (VarE 'S.upper, False)
+                "now" -> pure (VarE 'S.now, False)
+                "count" -> pure (VarE 'S.count, True)
+                "sum" -> pure (VarE 'S.sum_, True)
+                "avg" -> pure (VarE 'S.avg, True)
+                "min" -> pure (VarE 'S.min_, True)
+                "max" -> pure (VarE 'S.max_, True)
+                _ -> fail $ "Unsupported function: " <> fnNameStr
+
+            case maybeParams of
+              Nothing -> pure squealFn
+              Just params -> case params of
+                PGT_AST.NormalFuncApplicationParams maybeAllOrDistinct args maybeSortClause -> do
+                  when (isJust maybeSortClause) $
+                    fail "ORDER BY in function calls is not supported."
+                  argExps <- mapM (renderPGTFuncArgExpr cteNames) (NE.toList args)
+                  if isAggregate
+                    then do
+                      let
+                        aggArgConstructor = case maybeAllOrDistinct of
+                          Just True -> ConE 'S.Distincts
+                          _ -> ConE 'S.Alls
+                        npArgs = foldr (\h t -> ConE '(S.:*) `AppE` h `AppE` t) (ConE 'S.Nil) argExps
+                      pure $ squealFn `AppE` (aggArgConstructor `AppE` npArgs)
+                    else do
+                      when (isJust maybeAllOrDistinct) $
+                        fail "DISTINCT is not supported for non-aggregate functions."
+                      pure $ foldl' AppE squealFn argExps
+                PGT_AST.StarFuncApplicationParams ->
+                  if fnNameStrLower == "count"
+                    then pure $ VarE 'S.countStar
+                    else fail "Star argument only supported for COUNT"
+                _ -> fail $ "Unsupported function parameters structure: " <> show params
 
 
 renderPGTFuncArgExpr :: [Text.Text] -> PGT_AST.FuncArgExpr -> Q Exp
@@ -1325,7 +1653,8 @@ renderPGTFuncArgExpr cteNames = \case
   _ -> fail "Named or colon-syntax function arguments not supported"
 
 
-renderPGTFuncExprCommonSubexpr :: [Text.Text] -> PGT_AST.FuncExprCommonSubexpr -> Q Exp
+renderPGTFuncExprCommonSubexpr
+  :: [Text.Text] -> PGT_AST.FuncExprCommonSubexpr -> Q Exp
 renderPGTFuncExprCommonSubexpr cteNames = \case
   PGT_AST.CurrentTimestampFuncExprCommonSubexpr (Just _) ->
     fail "CURRENT_TIMESTAMP with precision is not supported."
