@@ -15,6 +15,7 @@ module Main (main) where
 
 import Data.Aeson (Value)
 import Data.Int (Int32, Int64)
+import Data.Scientific (Scientific)
 import Data.Text (Text)
 import Data.Time (Day, UTCTime)
 import Data.UUID (UUID)
@@ -532,6 +533,68 @@ main =
           squealRendering = "SELECT * FROM (VALUES (1, (E'text_val' :: text))) AS t (\"num\", \"txt\")"
         checkStatement squealRendering statement
 
+      it "values (1, 'a'), (2, 'b')" $ do
+        let
+          statement
+            :: Statement
+                 DB
+                 ()
+                 ( Field "_column1" Int64
+                 , ( Field "_column2" Text
+                   , ()
+                   )
+                 )
+          statement = [ssql| values (1, 'a'), (2, 'b') |]
+          squealRendering :: Text
+          squealRendering =
+            "SELECT * FROM (VALUES (1, (E'a' :: text)), (2, (E'b' :: text))) AS t (\"_column1\", \"_column2\")"
+        checkStatement squealRendering statement
+
+      it "values (1, 'a')" $ do
+        let
+          statement
+            :: Statement
+                 DB
+                 ()
+                 ( Field "_column1" Int64
+                 , ( Field "_column2" Text
+                   , ()
+                   )
+                 )
+          statement = [ssql| values (1, 'a') |]
+          squealRendering :: Text
+          squealRendering =
+            "SELECT * FROM (VALUES (1, (E'a' :: text))) AS t (\"_column1\", \"_column2\")"
+        checkStatement squealRendering statement
+
+      it "select * from (values (1, 'a'), (2, 'b')) as t (n, s)" $ do
+        let
+          statement
+            :: Statement
+                 DB
+                 ()
+                 ( Field "n" Int64
+                 , ( Field "s" Text
+                   , ()
+                   )
+                 )
+          statement =
+            [ssql| select * from (values (1, 'a'), (2, 'b')) as t (n, s) |]
+          squealRendering :: Text
+          squealRendering =
+            "SELECT * FROM (SELECT * FROM (VALUES (1, (E'a' :: text)), (2, (E'b' :: text))) AS t (\"n\", \"s\")) AS \"t\""
+        checkStatement squealRendering statement
+
+      -- Selecting FROM a subquery (VALUES ...) alias is not yet supported by
+      -- the translator (SelectTableRef). This can be added in a follow-up.
+
+
+      -- Negative tests (compile-time failures) cannot be included in the
+      -- test suite because the quasiquoter fails during compilation.
+
+      -- Selecting FROM a subquery (VALUES ...) alias is not yet supported by
+      -- the translator (SelectTableRef). This can be added in a follow-up.
+
       describe "group by" $ do
         it "select name from users group by name" $ do
           let
@@ -594,12 +657,62 @@ main =
                        )
                      )
                    )
-            statement = [ssql| with users_cte as (select * from users) select * from users_cte |]
+            statement =
+              [ssql|
+                with users_cte as (select * from users)
+                select * from users_cte
+              |]
             squealRendering :: Text
             squealRendering =
               "WITH \"users_cte\" AS (SELECT * FROM \"users\" AS \"users\") SELECT * FROM \"users_cte\" AS \"users_cte\""
           checkStatement squealRendering statement
 
+        it
+          "with recursive t as ( select 1 as n union all select (n + 1) as n from t where n < 100) select n from t"
+          $ do
+            let
+              statement
+                :: Statement
+                     DB
+                     ()
+                     (Field "n" Int64, ())
+              statement =
+                [ssql|
+                with recursive t as (
+                  select 1 as n
+                  union all
+                  select (n + 1) as n from t where n < 100
+                )
+                select n from t
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "WITH RECURSIVE \"t\" AS ((SELECT * FROM (VALUES (1)) AS t (\"n\")) UNION ALL (SELECT (\"n\" + 1) AS \"n\" FROM \"t\" AS \"t\" WHERE (\"n\" < 100))) SELECT \"n\" AS \"n\" FROM \"t\" AS \"t\""
+            checkStatement squealRendering statement
+
+        it
+          "with recursive users_cte as ( select id, name from users union all select id, name from users_cte) select * from users_cte"
+          $ do
+            let
+              statement
+                :: Statement
+                     DB
+                     ()
+                     (Field "id" Text, (Field "name" Text, ()))
+              statement =
+                [ssql|
+                with recursive users_cte as (
+                  select id, name from users
+                  union all
+                  select id, name from users_cte
+                )
+                select * from users_cte
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "WITH RECURSIVE \"users_cte\" AS ((SELECT \"id\" AS \"id\", \"name\" AS \"name\" FROM \"users\" AS \"users\") UNION ALL (SELECT \"id\" AS \"id\", \"name\" AS \"name\" FROM \"users_cte\" AS \"users_cte\")) SELECT * FROM \"users_cte\" AS \"users_cte\""
+            checkStatement squealRendering statement
+
         it
           "with users_cte as (select * from users), emails_cte as (select * from emails) select users_cte.*, emails_cte.email from users_cte join emails_cte on users_cte.id = emails_cte.user_id"
           $ do
@@ -633,6 +746,70 @@ main =
                 "WITH \"users_cte\" AS (SELECT * FROM \"users\" AS \"users\"), \"emails_cte\" AS (SELECT * FROM \"emails\" AS \"emails\") SELECT \"users_cte\".*, \"emails_cte\".\"email\" AS \"email\" FROM \"users_cte\" AS \"users_cte\" INNER JOIN \"emails_cte\" AS \"emails_cte\" ON (\"users_cte\".\"id\" = \"emails_cte\".\"user_id\")"
             checkStatement squealRendering statement
 
+      describe "set operations" $ do
+        it "select name from users union select name from users_copy" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| select name from users union select name from users_copy |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") UNION (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
+        it "select name from users union all select name from users_copy" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| select name from users union all select name from users_copy |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") UNION ALL (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
+        it "select name from users intersect select name from users_copy" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| select name from users intersect select name from users_copy |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") INTERSECT (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
+        it "select name from users intersect all select name from users_copy" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| select name from users intersect all select name from users_copy |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") INTERSECT ALL (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
+        it "select name from users except select name from users_copy" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| select name from users except select name from users_copy |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") EXCEPT (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
+        it "select name from users except all select name from users_copy" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| select name from users except all select name from users_copy |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") EXCEPT ALL (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
+        it "(select name from users) union (select name from users_copy)" $ do
+          let
+            statement :: Statement DB () (Field "name" Text, ())
+            statement = [ssql| (select name from users) union (select name from users_copy) |]
+            squealRendering :: Text
+            squealRendering =
+              "(SELECT \"name\" AS \"name\" FROM \"users\" AS \"users\") UNION (SELECT \"name\" AS \"name\" FROM \"users_copy\" AS \"users_copy\")"
+          checkStatement squealRendering statement
+
     describe "inserts" $ do
       it "insert into emails (id, user_id, email) values (1, 'user-1', 'foo@bar')" $ do
         let
@@ -941,6 +1118,103 @@ main =
                 "WITH \"new_user\" AS (SELECT * FROM (VALUES ((E'id_new' :: text), (E'new_name' :: text), (E'new_bio' :: text))) AS t (\"id\", \"name\", \"bio\")) INSERT INTO \"users_copy\" AS \"users_copy\" SELECT * FROM \"new_user\" AS \"new_user\""
             checkStatement squealRendering statement
 
+      describe "on conflict" $ do
+        it
+          "insert into users_copy (id, name, bio) values ('id1', 'name1', null) on conflict on constraint pk_users_copy do nothing"
+          $ do
+            let
+              statement :: Statement DB () ()
+              statement =
+                [ssql|
+                insert into users_copy (id, name, bio)
+                values ('id1', 'name1', null)
+                on conflict on constraint pk_users_copy do nothing
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "INSERT INTO \"users_copy\" AS \"users_copy\" (\"id\", \"name\", \"bio\") VALUES ((E'id1' :: text), (E'name1' :: text), NULL) ON CONFLICT ON CONSTRAINT \"pk_users_copy\" DO NOTHING"
+            checkStatement squealRendering statement
+
+        it
+          "insert into users_copy (id, name, bio) values ('id1', 'name1', 'bio1') on conflict on constraint pk_users_copy do update set name = 'new_name'"
+          $ do
+            let
+              statement :: Statement DB () ()
+              statement =
+                [ssql|
+                insert into users_copy (id, name, bio)
+                values ('id1', 'name1', 'bio1')
+                on conflict on constraint pk_users_copy
+                do update set name = 'new_name'
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "INSERT INTO \"users_copy\" AS \"users_copy\" (\"id\", \"name\", \"bio\") VALUES ((E'id1' :: text), (E'name1' :: text), (E'bio1' :: text)) ON CONFLICT ON CONSTRAINT \"pk_users_copy\" DO UPDATE SET \"name\" = (E'new_name' :: text)"
+            checkStatement squealRendering statement
+
+        it
+          "insert into users_copy (id, name, bio) values ('id1', 'name1', null) on conflict on constraint pk_users_copy do update set name = 'new_name' where users_copy.name = 'old_name'"
+          $ do
+            let
+              statement :: Statement DB () ()
+              statement =
+                [ssql|
+                insert into users_copy (id, name, bio)
+                values ('id1', 'name1', null)
+                on conflict on constraint pk_users_copy
+                do update set name = 'new_name'
+                where users_copy.name = 'old_name'
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "INSERT INTO \"users_copy\" AS \"users_copy\" (\"id\", \"name\", \"bio\") VALUES ((E'id1' :: text), (E'name1' :: text), NULL) ON CONFLICT ON CONSTRAINT \"pk_users_copy\" DO UPDATE SET \"name\" = (E'new_name' :: text) WHERE (\"users_copy\".\"name\" = (E'old_name' :: text))"
+            checkStatement squealRendering statement
+
+        it
+          "insert into users_copy (id, name, bio) values ('id1', 'name1', null) on conflict on constraint pk_users_copy do nothing returning id"
+          $ do
+            let
+              statement :: Statement DB () (Field "id" Text, ())
+              statement =
+                [ssql|
+                insert into users_copy (id, name, bio)
+                values ('id1', 'name1', null)
+                on conflict on constraint pk_users_copy do nothing
+                returning id
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "INSERT INTO \"users_copy\" AS \"users_copy\" (\"id\", \"name\", \"bio\") VALUES ((E'id1' :: text), (E'name1' :: text), NULL) ON CONFLICT ON CONSTRAINT \"pk_users_copy\" DO NOTHING RETURNING \"id\" AS \"id\""
+            checkStatement squealRendering statement
+
+        it
+          "insert into users_copy (id, name, bio) values ('id1', 'name1', 'bio1') on conflict on constraint pk_users_copy do update set name = 'new_name' returning *"
+          $ do
+            let
+              statement
+                :: Statement
+                     DB
+                     ()
+                     ( Field "id" Text
+                     , ( Field "name" Text
+                       , ( Field "bio" (Maybe Text)
+                         , ()
+                         )
+                       )
+                     )
+              statement =
+                [ssql|
+                insert into users_copy (id, name, bio)
+                values ('id1', 'name1', 'bio1')
+                on conflict on constraint pk_users_copy
+                do update set name = 'new_name'
+                returning *
+              |]
+              squealRendering :: Text
+              squealRendering =
+                "INSERT INTO \"users_copy\" AS \"users_copy\" (\"id\", \"name\", \"bio\") VALUES ((E'id1' :: text), (E'name1' :: text), (E'bio1' :: text)) ON CONFLICT ON CONSTRAINT \"pk_users_copy\" DO UPDATE SET \"name\" = (E'new_name' :: text) RETURNING *"
+            checkStatement squealRendering statement
+
     describe "deletes" $ do
       it "delete from users where true" $ do
         let
@@ -1376,6 +1650,85 @@ main =
             squealRendering = "SELECT * FROM (VALUES (CURRENT_DATE)) AS t (\"today\")"
           checkStatement squealRendering stmt
 
+        describe "aggregate functions" $ do
+          it "select sum(id) as total_ids from emails" $ do
+            let
+              stmt :: Statement DB () (Field "total_ids" (Maybe Int64), ())
+              stmt = [ssql| select sum(id) as total_ids from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT sum(ALL \"id\") AS \"total_ids\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select sum(all id) as total_ids from emails" $ do
+            let
+              stmt :: Statement DB () (Field "total_ids" (Maybe Int64), ())
+              stmt = [ssql| select sum(all id) as total_ids from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT sum(ALL \"id\") AS \"total_ids\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select sum(distinct id) as total_ids from emails" $ do
+            let
+              stmt :: Statement DB () (Field "total_ids" (Maybe Int64), ())
+              stmt = [ssql| select sum(distinct id) as total_ids from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT sum(DISTINCT \"id\") AS \"total_ids\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select count(distinct id) as distinct_ids from emails" $ do
+            let
+              stmt :: Statement DB () (Field "distinct_ids" Int64, ())
+              stmt = [ssql| select count(distinct id) as distinct_ids from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT count(DISTINCT \"id\") AS \"distinct_ids\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select count(all id) as all_ids from emails" $ do
+            let
+              stmt :: Statement DB () (Field "all_ids" Int64, ())
+              stmt = [ssql| select count(all id) as all_ids from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT count(ALL \"id\") AS \"all_ids\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select avg(id) as avg_id from emails" $ do
+            let
+              stmt
+                :: Statement
+                     DB
+                     ()
+                     ( Field "avg_id" (Maybe Scientific)
+                     , ()
+                     )
+              stmt = [ssql| select avg(id) as avg_id from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT avg(ALL \"id\") AS \"avg_id\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select min(id) as min_id from emails" $ do
+            let
+              stmt :: Statement DB () (Field "min_id" (Maybe Int32), ())
+              stmt = [ssql| select min(id) as min_id from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT min(ALL \"id\") AS \"min_id\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
+          it "select max(id) as max_id from emails" $ do
+            let
+              stmt :: Statement DB () (Field "max_id" (Maybe Int32), ())
+              stmt = [ssql| select max(id) as max_id from emails group by () |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT max(ALL \"id\") AS \"max_id\" FROM \"emails\" AS \"emails\""
+            checkStatement squealRendering stmt
+
         it "haskell variables in expressions" $ do
           let
             mkStatement
@@ -1517,6 +1870,71 @@ main =
           squealRendering = "SELECT * FROM \"users\" AS \"users\" FOR UPDATE"
         checkStatement squealRendering statement
 
+      it "select * from users for read only" $ do
+        let
+          statement
+            :: Statement
+                 DB
+                 ()
+                 ( Field "id" Text
+                 , ( Field "name" Text
+                   , ( Field "employee_id" UUID
+                     , ( Field "bio" (Maybe Text)
+                       , ()
+                       )
+                     )
+                   )
+                 )
+          statement = [ssql| select * from users for read only |]
+          squealRendering :: Text
+          squealRendering = "SELECT * FROM \"users\" AS \"users\""
+        checkStatement squealRendering statement
+
+      it "select * from users where name = 'bob' for read only" $ do
+        let
+          statement
+            :: Statement
+                 DB
+                 ()
+                 ( Field "id" Text
+                 , ( Field "name" Text
+                   , ( Field "employee_id" UUID
+                     , ( Field "bio" (Maybe Text)
+                       , ()
+                       )
+                     )
+                   )
+                 )
+          statement = [ssql| select * from users where name = 'bob' for read only |]
+          squealRendering :: Text
+          squealRendering = "SELECT * FROM \"users\" AS \"users\" WHERE (\"name\" = (E'bob' :: text))"
+        checkStatement squealRendering statement
+
+      it "with users_cte as (select * from users) select * from users_cte for read only" $ do
+        let
+          statement
+            :: Statement
+                 DB
+                 ()
+                 ( Field "id" Text
+                 , ( Field "name" Text
+                   , ( Field "employee_id" UUID
+                     , ( Field "bio" (Maybe Text)
+                       , ()
+                       )
+                     )
+                   )
+                 )
+          statement =
+            [ssql|
+              with users_cte as (select * from users)
+              select * from users_cte for read only
+            |]
+          squealRendering :: Text
+          squealRendering =
+            "WITH \"users_cte\" AS (SELECT * FROM \"users\" AS \"users\") SELECT * FROM \"users_cte\" AS \"users_cte\""
+        checkStatement squealRendering statement
+
       it "select * from jsonb_test" $ do
         let
           statement
@@ -1697,6 +2115,160 @@ main =
                 "SELECT \"employee_id\" AS \"employee_id\", count(ALL \"id\") AS \"_col2\" FROM \"users\" AS \"users\" GROUP BY \"employee_id\" HAVING (count(ALL \"id\") > 1)"
             checkStatement squealRendering statement
 
+      describe "window functions" $ do
+        it "select name, row_number() over () as rn from users" $ do
+          let
+            statement
+              :: Statement
+                   DB
+                   ()
+                   ( Field "name" Text
+                   , ( Field "rn" Int64
+                     , ()
+                     )
+                   )
+            statement = [ssql| select name, row_number() over () as rn from users |]
+            squealRendering :: Text
+            squealRendering =
+              "SELECT \"name\" AS \"name\", row_number() OVER () AS \"rn\" FROM \"users\" AS \"users\""
+          checkStatement squealRendering statement
+
+        it
+          "select name, rank() over (partition by employee_id order by name) as r from users"
+          $ do
+            let
+              statement
+                :: Statement
+                     DB
+                     ()
+                     ( Field "name" Text
+                     , ( Field "r" Int64
+                       , ()
+                       )
+                     )
+              statement =
+                [ssql| select name, rank() over (partition by employee_id order by name) as r from users |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT \"name\" AS \"name\", rank() OVER (PARTITION BY \"employee_id\" ORDER BY \"name\" ASC) AS \"r\" FROM \"users\" AS \"users\""
+            checkStatement squealRendering statement
+
+        it "select email, sum(id) over (partition by user_id) as user_total from emails" $ do
+          let
+            statement
+              :: Statement
+                   DB
+                   ()
+                   ( Field "email" (Maybe Text)
+                   , ( Field "user_total" (Maybe Int64)
+                     , ()
+                     )
+                   )
+            statement =
+              [ssql| select email, sum(id) over (partition by user_id) as user_total from emails |]
+            squealRendering :: Text
+            squealRendering =
+              "SELECT \"email\" AS \"email\", sum(\"id\") OVER (PARTITION BY \"user_id\") AS \"user_total\" FROM \"emails\" AS \"emails\""
+          checkStatement squealRendering statement
+
+        it "select email, avg(id) over (partition by user_id) as user_avg from emails" $ do
+          let
+            statement
+              :: Statement
+                   DB
+                   ()
+                   ( Field "email" (Maybe Text)
+                   , ( Field "user_avg" (Maybe Scientific)
+                     , ()
+                     )
+                   )
+            statement =
+              [ssql| select email, avg(id) over (partition by user_id) as user_avg from emails |]
+            squealRendering :: Text
+            squealRendering =
+              "SELECT \"email\" AS \"email\", avg(\"id\") OVER (PARTITION BY \"user_id\") AS \"user_avg\" FROM \"emails\" AS \"emails\""
+          checkStatement squealRendering statement
+
+        it "select email, min(id) over (partition by user_id) as user_min from emails" $ do
+          let
+            statement
+              :: Statement
+                   DB
+                   ()
+                   ( Field "email" (Maybe Text)
+                   , ( Field "user_min" (Maybe Int32)
+                     , ()
+                     )
+                   )
+            statement =
+              [ssql| select email, min(id) over (partition by user_id) as user_min from emails |]
+            squealRendering :: Text
+            squealRendering =
+              "SELECT \"email\" AS \"email\", min(\"id\") OVER (PARTITION BY \"user_id\") AS \"user_min\" FROM \"emails\" AS \"emails\""
+          checkStatement squealRendering statement
+
+        it "select email, max(id) over (partition by user_id) as user_max from emails" $ do
+          let
+            statement
+              :: Statement
+                   DB
+                   ()
+                   ( Field "email" (Maybe Text)
+                   , ( Field "user_max" (Maybe Int32)
+                     , ()
+                     )
+                   )
+            statement =
+              [ssql| select email, max(id) over (partition by user_id) as user_max from emails |]
+            squealRendering :: Text
+            squealRendering =
+              "SELECT \"email\" AS \"email\", max(\"id\") OVER (PARTITION BY \"user_id\") AS \"user_max\" FROM \"emails\" AS \"emails\""
+          checkStatement squealRendering statement
+
+        it
+          "select name, row_number() over (order by name), rank() over (order by name) from users"
+          $ do
+            let
+              statement
+                :: Statement
+                     DB
+                     ()
+                     ( Field "name" Text
+                     , ( Field "_window1" Int64
+                       , ( Field "_window2" Int64
+                         , ()
+                         )
+                       )
+                     )
+              statement =
+                [ssql| select name, row_number() over (order by name), rank() over (order by name) from users |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT \"name\" AS \"name\", row_number() OVER ( ORDER BY \"name\" ASC) AS \"_window1\", rank() OVER ( ORDER BY \"name\" ASC) AS \"_window2\" FROM \"users\" AS \"users\""
+            checkStatement squealRendering statement
+
+        it
+          "select name, row_number() over (partition by employee_id order by name), rank() over (order by name) from users"
+          $ do
+            let
+              statement
+                :: Statement
+                     DB
+                     ()
+                     ( Field "name" Text
+                     , ( Field "_window1" Int64
+                       , ( Field "_window2" Int64
+                         , ()
+                         )
+                       )
+                     )
+              statement =
+                [ssql| select name, row_number() over (partition by employee_id order by name), rank() over (order by name) from users |]
+              squealRendering :: Text
+              squealRendering =
+                "SELECT \"name\" AS \"name\", row_number() OVER (PARTITION BY \"employee_id\" ORDER BY \"name\" ASC) AS \"_window1\", rank() OVER ( ORDER BY \"name\" ASC) AS \"_window2\" FROM \"users\" AS \"users\""
+            checkStatement squealRendering statement
+
 
 _printQuery :: (RenderSQL a) => a -> IO ()
 _printQuery = putStrLn . T.unpack . TE.decodeUtf8 . renderSQL
